{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE PackageImports #-}
{-# LANGUAGE RecordWildCards #-}
{-# LANGUAGE GeneralizedNewtypeDeriving #-}

{-# LANGUAGE NoImplicitPrelude #-}

import Relude

import "cryptonite" Crypto.Random (MonadRandom(..))

import Control.Monad.Logger (runStderrLoggingT)
import Data.Aeson (toJSON)
import Database.Persist.Postgresql (withPostgresqlConn, runSqlConn)
import System.IO.Echo (withoutInputEcho)

import Options.Applicative

import Datarekisteri.Backend.Sql (MonadSql, runQuery)

import qualified Datarekisteri.Backend.Sql.Queries as Sql

import Datarekisteri.Backend.Sql.Types
import Datarekisteri.Backend.Types
import Datarekisteri.Backend.Utils
import Datarekisteri.Core.Types

main :: IO ()
main = do
    CLIOptions {..} <- execParser $ info (cliOptions <**> helper) mempty
    runCLIM optionsDBUrl $ case optionsSubCommand of
      (AddUser addUserOpts) -> addUserMain addUserOpts
      GCEmails -> gcEmailsMain
      GCApplications -> gcApplicationsMain
      GCAll -> gcAllMain

addUserMain :: AddUserOpts -> CLIM ()
addUserMain AddUserOpts {..} = do
    time <- currentTime
    passwordHash <- putStr "Password: " >> hFlush stdout >> liftIO (withoutInputEcho getLine) >>= hashPassword
    runQuery $ do
        userID <- Sql.addUser $ SqlUser
            { sqlUserRegistered = time
            , sqlUserPasswordCrypt = passwordHash
            , sqlUserPermissions = show addUserPermissions
            , sqlUserAccepted = Just time
            , sqlUserMemberData = toJSON $ MemberData
                { nickname = addUserNickname
                , name = addUserName
                , birthdate = addUserBirthdate
                , homeplace = addUserHomeplace
                , application = addUserApplication
                , phoneNumber = addUserPhoneNumber
                }
            }
        void $ Sql.addEmail $ SqlEmail
            { sqlEmailUid = fromID userID
            , sqlEmailEmail = addUserEmail
            , sqlEmailVid = Nothing
            }

gcEmailsMain :: CLIM ()
gcEmailsMain = do
    time <- currentTime
    runQuery $ do
        Sql.deleteExpiredEmails time
        Sql.deleteOrphanedVerifications

gcApplicationsMain :: CLIM ()
gcApplicationsMain = runQuery $ Sql.deleteUsersWithoutEmail

gcAllMain :: CLIM ()
gcAllMain = do
    gcEmailsMain
    gcApplicationsMain

cliOptions :: Parser CLIOptions
cliOptions = CLIOptions
    <$> strOption (short 'u' <> long "db-url" <> metavar "URL" <> value "postgres:///datarekisteri-backend")
    <*> cliCommandParser

cliCommandParser :: Parser CLISubCommand
cliCommandParser = hsubparser
    $ command "add-user" (info addUserCommand (progDesc "Add a user to datarekisteri"))
    <> command "gc-expired-emails" (info (pure GCEmails) (progDesc "Delete emails that have not been verified in time."))
    <> command "gc-applications" (info (pure GCApplications) (progDesc "Delete users without an email address"))
    <> command "gc" (info (pure GCAll) (progDesc "Run all garbage collection jobs."))

addUserCommand :: Parser CLISubCommand
addUserCommand = fmap AddUser $ AddUserOpts
    <$> optional (strOption (long "nickname" <> metavar "NAME"))
    <*> strOption (long "name" <> metavar "NAME")
    <*> option (maybeReader $ toDate . toText) (long "birthdate" <> metavar "DATE" <> help "The user's birthdate, YYYY-MM-DD")
    <*> strOption (long "homeplace" <> metavar "NAME" <> help "The user's homeplace, usually a city")
    <*> option (maybeReader $ toPhoneNumber . toText) (long "phone-number" <> metavar "PHONE" <> help "The user's phone number. Only numbers, spaces and the plus-sign are allowed")
    <*> option (maybeReader $ toEmail . toText) (long "email" <> metavar "EMAIL" <> help "The user's email address.")
    <*> (fromList <$> many permissionParser)
    <*> strOption (long "application" <> metavar "TEXT" <> value "Added by the admin.")

permissionParser :: Parser (Scope, Permission)
permissionParser = (,)
    <$> option auto (long "scope")
    <*> option auto (long "permission")

data CLIOptions = CLIOptions 
    { optionsDBUrl :: String
    , optionsSubCommand :: CLISubCommand
    }

data CLISubCommand = AddUser AddUserOpts
                   | GCEmails
                   | GCApplications
                   | GCAll

data AddUserOpts = AddUserOpts
        { addUserNickname :: Maybe Text
        , addUserName :: Text
        , addUserBirthdate :: Date
        , addUserHomeplace :: Text
        , addUserPhoneNumber :: PhoneNumber
        , addUserEmail :: Email
        , addUserPermissions :: Map Scope Permission
        , addUserApplication :: Text
        }

newtype CLIM a = CLIM (ReaderT String IO a)
    deriving (Functor, Applicative, Monad, MonadIO, MonadReader String)

instance MonadTime CLIM where
    currentTime = liftIO currentTime

instance MonadSql CLIM where
    runQuery query = do
        dbUrl <- ask
        liftIO $ runStderrLoggingT $ withPostgresqlConn (encodeUtf8 dbUrl) $ runSqlConn query

instance MonadRandom CLIM where
    getRandomBytes = liftIO . getRandomBytes

runCLIM :: String -> CLIM a -> IO a
runCLIM dbUrl (CLIM m) = runReaderT m dbUrl
